/*
    scconfig - test code for default and scripts
    Copyright (C) 2009  Tibor Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

		Project page: http://repo.hu/projects/scconfig
		Contact via email: scconfig [at] igor2.repo.hu
*/

#include <stdlib.h>
#include <string.h>
#include "db.h"
#include "find.h"
#include "log.h"
#include "arg.h"
#include "dep.h"
#include "deps_default.h"
#include "libs.h"
#include "hooks.h"

void re_fail(char *s, char c)
{
	fprintf(stderr, "Regex error: %s [opcode %o]\n", s, c);
	abort();
}

#define MAX_CUSTOM_REQS 32
char *custom_reqs[MAX_CUSTOM_REQS];
int num_custom_reqs = 0;

void custom_arg(const char *key, const char *value)
{
	if (hook_custom_arg(key, value))
		return;
	if (strcmp(key, "detect") == 0) {
		printf("Will detect: %s\n", value);
		if (num_custom_reqs >= MAX_CUSTOM_REQS) {
			report("Too many custom reqs from the command line, exiting\n");
			exit(1);
		}
		custom_reqs[num_custom_reqs] = strclone(value);
		num_custom_reqs++;
		return;
	}
	error("Unknown argument '%s'\n", key);
}

#ifdef PLUGIN_SCRIPTS
#include "../scripts/INIT.h"
#endif

#ifdef PLUGIN_PARSER
#include "../parser/INIT.h"
#endif

#ifdef PLUGIN_C99
#include "../c99/INIT.h"
#endif

#ifdef PLUGIN_PARSGEN
#include "../parsgen/INIT.h"
#endif

#ifdef PLUGIN_MATH
#include "../math/INIT.h"
#endif

#ifdef PLUGIN_SOCKET
#include "../socket/INIT.h"
#endif

#ifdef PLUGIN_GENERATOR
#include "generator.h"
#endif



void init(void)
{
	db_init();
	log_init();
	dep_init();
	deps_default_init();

	/* common internal directory */
	db_mkdir("/internal");

	/* We have a host system for sure - also make it the default */
	db_mkdir("/host");
	db_cd("/host");

	/* emulator for the host system is empty string */
	put("sys/emu", "");

#ifdef PLUGIN_SCRIPTS
#include "../scripts/INIT.c"
#endif

#ifdef PLUGIN_PARSER
#include "../parser/INIT.c"
#endif

#ifdef PLUGIN_C99
#include "../c99/INIT.c"
#endif

#ifdef PLUGIN_PARSGEN
#include "../parsgen/INIT.c"
#endif

#ifdef PLUGIN_MATH
#include "../math/INIT.c"
#endif

#ifdef PLUGIN_SOCKET
#include "../socket/INIT.c"
#endif

#ifdef PLUGIN_GENERATOR
#include "../generator/INIT.c"
#endif
}

void uninit(void)
{
	log_uninit();
	dep_uninit();
	db_uninit();
}

void run_custom_reqs(void)
{
	int n;
	if (num_custom_reqs > 0) {
		printf("Running custom requirements\n");
		for(n = 0; n < num_custom_reqs; n++)
			require(custom_reqs[n], 1, 1);
	}
}

int no_autodetect_sys = 0;
int no_save_cache = 0;
int main(int argc, char *argv[])
{
	re_modw("./\\");
	if (hook_preinit()) {
		fprintf(stderr, "hook_preinit failed, exiting\n");
		return 1;
	}
	init();
	if (hook_postinit())  {
		fprintf(stderr, "hook_postinit failed, exiting\n");
		return 1;
	}

	process_args(argc, argv);
	if (hook_postarg()) {
		fprintf(stderr, "hook_postarg failed, exiting\n");
		return 1;
	}

	if (!no_autodetect_sys) {
		find_target(0, 1);
		printf("--- Detecting host\n");

		require("sys/name", 0, 1);
	}

	run_custom_reqs();
	if (hook_detect_host())  {
		fprintf(stderr, "hook_detect_host failed, exiting\n");
		return 1;
	}

	if (!no_autodetect_sys) {
		if (!iscross)
			printf("--- Detecting target (same as host)\n");
		else
			printf("--- Detecting target (differs from host)\n");
	}
	db_cd("/target");
	run_custom_reqs();


	if (hook_detect_target()) {
		fprintf(stderr, "hook_detect_target failed, exiting\n");
		return 1;
	}

	if (hook_generate()) {
		fprintf(stderr, "hook_generate failed, exiting\n");
		return 1;
	}

	if (!no_save_cache)
		export("config.cache", 1, "/");

	hook_preuninit();
	uninit();
	hook_postuninit();
	return 0;
}

