
/********************************************
fin.c
copyright 1991, 1992.  Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

/*$Log: fin.c,v $
 * Revision 1.10  1995/12/24  22:23:22  mike
 * remove mawk_errmsg() from inside mawk_FINopen
 *
 * Revision 1.9  1995/06/06  00:18:29  mike
 * change mawk_exit(1) to mawk_exit(2)
 *
 * Revision 1.8  1994/12/13  00:26:35  mike
 * rt_nr and rt_fnr for run-time error messages
 *
 * Revision 1.7  1994/12/11  23:25:05  mike
 * -Wi option
 *
 * Revision 1.6  1994/12/11  22:14:15  mike
 * remove THINK_C #defines.  Not a political statement, just no indication
 * that anyone ever used it.
 *
 * Revision 1.5  1994/10/08  19:15:42  mike
 * remove SM_DOS
 *
 * Revision 1.4  1993/07/17  13:22:55  mike
 * indent and general code cleanup
 *
 * Revision 1.3  1993/07/15  13:26:55  mike
 * SIZE_T and indent
 *
 * Revision 1.2	 1993/07/04  12:51:57  mike
 * start on autoconfig changes
 *
 * Revision 1.1.1.1  1993/07/03	 18:58:13  mike
 * move source to cvs
 *
 * Revision 5.7	 1993/01/01  21:30:48  mike
 * split mawk_new_STRING() into mawk_new_STRING and mawk_new_STRING0
 *
 * Revision 5.6	 1992/12/17  02:48:01  mike
 * 1.1.2d changes for DOS
 *
 * Revision 5.5	 1992/07/28  15:11:30  brennan
 * minor change in mawk_finding eol, needed for MsDOS
 *
 * Revision 5.4	 1992/07/10  16:17:10  brennan
 * MsDOS: remove NO_BINMODE macro
 *
 * Revision 5.3	 1992/07/08  16:14:27  brennan
 * FILENAME and FNR retain last values in the
 * END block.
 *
 * Revision 5.2	 1992/02/21  13:30:08  brennan
 * fixed bug that free'd FILENAME twice if
 * command line was var=value only
 *
 * Revision 5.1	 91/12/05  07:56:02  brennan
 * 1.1 pre-release
 *
*/

/* fin.c */

#include "mawk.h"
#include "fin.h"
#include "memory.h"
#include "bi_vars.h"
#include "field.h"
#include "symtype.h"
#include "scan.h"

#ifndef	  NO_FCNTL_H
#include <fcntl.h>
#endif

/* This file handles input files.  Opening, closing,
   buffering and (most important) splitting files into
   records, mawk_FINgets().
*/

int PROTO(isatty, (int));

static FIN *PROTO(next_main, (mawk_state_t *, int));
static char *PROTO(enlarge_fin_buffer, (mawk_state_t *, FIN *));
static void PROTO(set_main_to_stdin, (mawk_state_t *));
int PROTO(mawk_is_cmdline_assign, (mawk_state_t *, char *));	/* also used by init */

/* convert file-descriptor to FIN*.
   It's the main stream if main_flag is set
*/
FIN *mawk_FINdopen(MAWK, fd, main_flag)
		 mawk_state_t *MAWK;
		 int fd, main_flag;
{
	register FIN *fin = ZMALLOC(FIN);

	fin->fd = fd;
	fin->flags = main_flag ? (MAIN_FLAG | START_FLAG) : START_FLAG;
	fin->buffp = fin->buff = (char *) zmalloc(MAWK, BUFFSZ + 1);
	fin->nbuffs = 1;
	fin->buff[0] = 0;
	fin->artificial_size = 0;
	fin->artificial_buff = NULL;

	if (fd < 0) {
		fin->fp = (FILE *) 0;
		return fin;
	}

	if (isatty(fd) && MAWK->rs_shadow.type == SEP_CHAR && MAWK->rs_shadow.c == '\n' || MAWK->interactive_flag && fd == 0) {
		/* interactive, i.e., line buffer this file */
		if (fd == 0)
			fin->fp = stdin;
		else if (!(fin->fp = fdopen(fd, "r"))) {
			mawk_errmsg(MAWK, errno, "fdopen failed");
			mawk_exitval(MAWK, 2, NULL);
		}
	}
	else
		fin->fp = (FILE *) 0;

	return fin;
}

/* open a FIN* by filename.
   It's the main stream if main_flag is set.
   Recognizes "-" as stdin.
*/

FIN *mawk_FINopen(MAWK, filename, main_flag)
		 mawk_state_t *MAWK;
		 char *filename;
		 int main_flag;
{
	int fd;
	int oflag = O_RDONLY;

#if  MSDOS
	int bm = mawk_binmode() & 1;
	if (bm)
		oflag |= O_BINARY;
#endif

	if (filename[0] == '-' && filename[1] == 0) {
#if  MSDOS
		if (bm)
			setmode(0, O_BINARY);
#endif
		return mawk_FINdopen(MAWK, 0, main_flag);
	}

	if ((fd = open(filename, oflag, 0)) == -1)
		return (FIN *) 0;
	else
		return mawk_FINdopen(MAWK, fd, main_flag);
}

/* frees the buffer and fd, but leaves FIN structure until
   the user calls close() */

void mawk_FINsemi_close(MAWK, fin)
		 mawk_state_t *MAWK;
		 register FIN *fin;
{
	static char dead = 0;

	if (fin->buff != &dead) {
		zfree(MAWK, fin->buff, fin->nbuffs * BUFFSZ + 1);

		if (fin->fd)
			if (fin->fp)
				fclose(fin->fp);
			else
				close(fin->fd);

		fin->buff = fin->buffp = &dead;	/* marks it semi_closed */
	}
	/* else was already semi_closed */
}

/* user called close() on input file */
void mawk_FINclose(MAWK, fin)
		 mawk_state_t *MAWK;
		 FIN *fin;
{
	mawk_FINsemi_close(MAWK, fin);
	ZFREE(fin);
}

/* return one input record as determined by RS,
   from input file (FIN)  fin
*/

char *mawk_FINgets(MAWK, fin, len_p)
		 mawk_state_t *MAWK;
		 FIN *fin;
		 unsigned *len_p;
{
	register char *p, *q;
	unsigned match_len;
	unsigned r;

restart:

	if (!(p = fin->buffp)[0]) {		/* need a refill */
		if (fin->flags & EOF_FLAG) {
			if (fin->flags & MAIN_FLAG) {
				fin = next_main(MAWK, 0);
				goto restart;
			}
			else {
				*len_p = 0;
				return (char *) 0;
			}
		}

		if (fin->fp) {
			/* line buffering */
			if (!fgets(fin->buff, BUFFSZ + 1, fin->fp)) {
				fin->flags |= EOF_FLAG;
				fin->buff[0] = 0;
				fin->buffp = fin->buff;
				goto restart;						/* might be main_fin */
			}
			else {										/* return this line */

				/* mawk_find eol */
				p = fin->buff;
				while (*p != '\n' && *p != 0)
					p++;

				*p = 0;
				*len_p = p - fin->buff;
				fin->buffp = p;
				return fin->buff;
			}
		}
		else {
			/* block buffering */
			r = mawk_fillbuff(MAWK, fin->fd, fin->buff, fin->nbuffs * BUFFSZ, fin);
			if (r == 0) {
				fin->flags |= EOF_FLAG;
				fin->buffp = fin->buff;
				goto restart;						/* might be main */
			}
			else if (r == mawk_FIN_nomore)
				return mawk_FIN_nomore;

/* we will detect eof next mawk_FINgets anyway
			else if (r < fin->nbuffs * BUFFSZ) {
				fin->flags |= EOF_FLAG;
			}*/

			p = fin->buffp = fin->buff;

			if (fin->flags & START_FLAG) {
				fin->flags &= ~START_FLAG;
				if (MAWK->rs_shadow.type == SEP_MLR) {
					/* trim blank lines from front of file */
					while (*p == '\n')
						p++;
					fin->buffp = p;
					if (*p == 0)
						goto restart;
				}
			}
		}
	}

retry:

	switch (MAWK->rs_shadow.type) {
	case SEP_CHAR:
		q = strchr(p, MAWK->rs_shadow.c);
		match_len = 1;
		break;

	case SEP_STR:
		q = mawk_str_str(p, ((STRING *) MAWK->rs_shadow.ptr)->str, match_len = ((STRING *) MAWK->rs_shadow.ptr)->len);
		break;

	case SEP_MLR:
	case SEP_RE:
		q = mawk_re_pos_match(MAWK, p, MAWK->rs_shadow.ptr, &match_len);
		/* if the match is at the end, there might still be
		   more to match in the file */
		if (q && q[match_len] == 0 && !(fin->flags & EOF_FLAG))
			q = (char *) 0;
		break;

	default:
		mawk_bozo(MAWK, "type of rs_shadow");
	}

	if (q) {
		/* the easy and normal case */
		*q = 0;
		*len_p = q - p;
		fin->buffp = q + match_len;
		return p;
	}

	if (fin->flags & EOF_FLAG) {
		/* last line without a record terminator */
		*len_p = r = strlen(p);
		fin->buffp = p + r;

		if (MAWK->rs_shadow.type == SEP_MLR && fin->buffp[-1] == '\n' && r != 0) {
			(*len_p)--;
			*--fin->buffp = 0;
		}
		return p;
	}

	if (p == fin->buff) {
		/* current record is too big for the input buffer, grow buffer */
		p = enlarge_fin_buffer(MAWK, fin);
	}
	else {
		/* move a partial line to front of buffer and try again */
		unsigned rr;

		p = (char *) memcpy(fin->buff, p, r = strlen(p));
		q = p + r;
		rr = fin->nbuffs * BUFFSZ - r;

		if ((r = mawk_fillbuff(MAWK, fin->fd, q, rr, fin)) < rr) {
			if (r == mawk_FIN_nomore)
				return mawk_FIN_nomore;
			fin->flags |= EOF_FLAG;
		}
	}
	goto retry;
}

static char *enlarge_fin_buffer(MAWK, fin)
		 mawk_state_t *MAWK;
		 FIN *fin;
{
	unsigned r;
	unsigned oldsize = fin->nbuffs * BUFFSZ + 1;

#ifdef  MSDOS
	/* I'm not sure this can really happen:
	   avoid "16bit wrap" */
	if (fin->nbuffs == MAX_BUFFS) {
		mawk_errmsg(MAWK, 0, "out of input buffer space");
		mawk_exit(MAWK, 2);
	}
#endif

	fin->buffp = fin->buff = (char *) mawk_zrealloc(MAWK, fin->buff, oldsize, oldsize + BUFFSZ);
	fin->nbuffs++;

	r = mawk_fillbuff(MAWK, fin->fd, fin->buff + (oldsize - 1), BUFFSZ, fin);
	if (r < BUFFSZ)
		fin->flags |= EOF_FLAG;

	return fin->buff;
}

/*--------
  target is big enough to hold size + 1 chars
  on exit the back of the target is zero terminated
 *--------------*/
unsigned mawk_fillbuff(MAWK, fd, target, size, fin)
		 mawk_state_t *MAWK;
		 int fd;
		 register char *target;
		 unsigned size;
		 FIN *fin;
{
	register int r;
	unsigned entry_size = size;

	if ((fin != NULL) && (fin->artificial_size > 0)) {
		if (size < fin->artificial_size) { /* read part of the input */
			memcpy(target, fin->artificial_buff, size);
			target += size;
			fin->artificial_size = fin->artificial_size - size;
			memmove(fin->artificial_buff, fin->artificial_buff + size, fin->artificial_size);
			fin->artificial_size -= size;
			size = 0;
		}
		else { /* size >= artificial_size, we can read all input */
			memcpy(target, fin->artificial_buff, fin->artificial_size);
			target += fin->artificial_size;
			free(fin->artificial_buff);
			size -= fin->artificial_size;
			fin->artificial_buff = NULL;
			fin->artificial_size = 0;
		}
	}
	else if ((fin != NULL) && (fin->fd == -2)) {
		return mawk_FIN_nomore;
	}
	else {
		while (size)
			switch (r = read(fd, target, size)) {
			case -1:
				mawk_errmsg(MAWK, errno, "read error");
				mawk_exitval(MAWK, 2, -1);

			case 0:
				goto out;

			default:
				target += r;
				size -= r;
				break;
			}
	}
out:
	*target = 0;
	return entry_size - size;
}

/* main_fin is a handle to the main input stream
   == 0	 never been opened   */

static void set_main_to_stdin(mawk_state_t * MAWK)
{
	cell_destroy(FILENAME);
	FILENAME->type = C_STRING;
	FILENAME->ptr = (PTR) mawk_new_STRING(MAWK, "-");
	cell_destroy(FNR);
	FNR->type = C_DOUBLE;
	FNR->dval = 0.0;
	MAWK->rt_fnr = 0;
	MAWK->main_fin = mawk_FINdopen(MAWK, 0, 1);
}


/* this gets called once to get the input stream going.
   It is called after the execution of the BEGIN block
   unless there is a getline inside BEGIN {}
*/
void mawk_open_main(mawk_state_t * MAWK)
{
	CELL argc;

#if  MSDOS
	int k = mawk_binmode();

	if (k & 1)
		setmode(0, O_BINARY);
	if (k & 2) {
		setmode(1, O_BINARY);
		setmode(2, O_BINARY);
	}
#endif

	mawk_cellcpy(MAWK, &argc, ARGC);
	if (argc.type != C_DOUBLE)
		mawk_cast1_to_d(MAWK, &argc);

	if (argc.dval == 1.0)
		set_main_to_stdin(MAWK);
	else
		next_main(MAWK, 1);
}

/* get the next command line file open */
static FIN *next_main(MAWK, open_flag)
		 mawk_state_t *MAWK;
		 int open_flag;							/* called by mawk_open_main() if on */
{
	register CELL *cp;
	CELL argc;										/* copy of ARGC */
	CELL c_argi;									/* cell copy of argi */
	CELL argval;									/* copy of ARGV[c_argi] */


	argval.type = C_NOINIT;
	c_argi.type = C_DOUBLE;

	if (MAWK->main_fin)
		mawk_FINclose(MAWK, MAWK->main_fin);
	/* FILENAME and FNR don't change unless we really open
	   a new file */

	/* make a copy of ARGC to avoid side effect */
	if (mawk_cellcpy(MAWK, &argc, ARGC)->type != C_DOUBLE)
		mawk_cast1_to_d(MAWK, &argc);

	while (MAWK->argi < argc.dval) {
		c_argi.dval = MAWK->argi;
		MAWK->argi += 1.0;

		if (!(cp = mawk_array_find(MAWK, MAWK->Argv, &c_argi, NO_CREATE)))
			continue;									/* its deleted */

		/* make a copy so we can mawk_cast w/o side effect */
		cell_destroy(&argval);
		cp = mawk_cellcpy(MAWK, &argval, cp);
		if (cp->type < C_STRING)
			mawk_cast1_to_s(MAWK, cp);
		if (string(cp)->len == 0)
			continue;
		/* file argument is "" */

		/* it might be a command line assignment */
		if (mawk_is_cmdline_assign(MAWK, string(cp)->str))
			continue;

		/* try to open it -- we used to continue on failure,
		   but posix says we should quit */
		if (!(MAWK->main_fin = mawk_FINopen(MAWK, string(cp)->str, 1))) {
			mawk_errmsg(MAWK, errno, "cannot open %s", string(cp)->str);
			mawk_exitval(MAWK, 2, NULL);
		}

		/* success -- set FILENAME and FNR */
		cell_destroy(FILENAME);
		mawk_cellcpy(MAWK, FILENAME, cp);
		free_STRING(string(cp));
		cell_destroy(FNR);
		FNR->type = C_DOUBLE;
		FNR->dval = 0.0;
		MAWK->rt_fnr = 0;

		return MAWK->main_fin;
	}
	/* failure */
	cell_destroy(&argval);

	if (open_flag) {
		/* all arguments were null or assignment */
		set_main_to_stdin(MAWK);
		return MAWK->main_fin;
	}

	/* real failure */
	{
		/* this is how we mark EOF on main_fin  */
		static char dead_buff = 0;
		static FIN dead_main = { 0, (FILE *) 0, &dead_buff, &dead_buff,
			1, EOF_FLAG
		};

		return MAWK->main_fin = &dead_main;
		/* since MAIN_FLAG is not set, mawk_FINgets won't call next_main() */
	}
}


int mawk_is_cmdline_assign(MAWK, s)
		 mawk_state_t *MAWK;
		 char *s;
{
	register char *p;
	int c;
	SYMTAB *stp;
	CELL *cp;
	unsigned len;
	CELL cell;										/* used if command line assign to pseudo field */
	CELL *fp = (CELL *) 0;				/* ditto */

	if (MAWK->scan_code[*(unsigned char *) s] != SC_IDCHAR)
		return 0;

	p = s + 1;
	while ((c = MAWK->scan_code[*(unsigned char *) p]) == SC_IDCHAR || c == SC_DIGIT)
		p++;

	if (*p != '=')
		return 0;

	*p = 0;
	stp = mawk_find(MAWK, s, 1);

	switch (stp->type) {
	case ST_NONE:
		stp->type = ST_VAR;
		stp->stval.cp = cp = ZMALLOC(CELL);
		break;

	case ST_VAR:
	case ST_NR:									/* !! no one will do this */
		cp = stp->stval.cp;
		cell_destroy(cp);
		break;

	case ST_FIELD:
		/* must be pseudo field */
		fp = stp->stval.cp;
		cp = &cell;
		break;

	default:
		mawk_rt_error(MAWK, "cannot command line assign to %s\n\ttype clash or keyword", s);
	}

	/* we need to keep ARGV[i] intact */
	*p++ = '=';
	len = strlen(p) + 1;
	/* posix says escape sequences are on from command line */
	p = mawk_rm_escape(MAWK, strcpy((char *) zmalloc(MAWK, len), p));
	cp->ptr = (PTR) mawk_new_STRING(MAWK, p);
	zfree(MAWK, p, len);
	mawk_check_strnum(MAWK, cp);	/* sets cp->type */
	if (fp) {											/* move it from cell to pfield[] */
		mawk_field_assign(MAWK, fp, cp);
		free_STRING(string(cp));
	}
	return 1;
}
